"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import logging
from urllib.parse import urlparse
from opentelemetry.sdk.trace import Span
from okahu_apptrace.instrumentation.common.utils import (
    Option,
    get_keys_as_tuple,
    get_nested_value,
    try_option,
)

logger = logging.getLogger(__name__)


def extract_messages(args):
    """Extract system and user messages"""
    try:
        messages = []
        if args and isinstance(args, (list, tuple)) and args[0]:
            for msg in args[0]:
                if hasattr(msg, 'content') and hasattr(msg, 'role'):
                    role = getattr(msg.role, 'value', msg.role)
                    if role == "system":
                        messages.append({role: msg.content})
                    elif role in ["user", "human"]:
                        user_message = extract_query_from_content(msg.content)
                        messages.append({role: user_message})
        if args and isinstance(args, dict):
            for msg in args.get("messages", []):
                if hasattr(msg, 'content') and hasattr(msg, 'role'):
                    role = getattr(msg.role, 'value', msg.role)
                    if role == "system":
                        messages.append({role: msg.content})
                    elif role in ["user", "human"]:
                        user_message = msg.content
                        messages.append({role: user_message})
        return [str(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []


def extract_assistant_message(response):
    try:
        if isinstance(response, str):
            return [response]
        if hasattr(response, "content"):
            return [response.content]
        if hasattr(response, "message") and hasattr(response.message, "content"):
            return [response.message.content]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_assistant_message: %s", str(e))
        return []


def extract_query_from_content(content):
    try:
        query_prefix = "Query:"
        answer_prefix = "Answer:"
        query_start = content.find(query_prefix)
        if query_start == -1:
            return None

        query_start += len(query_prefix)
        answer_start = content.find(answer_prefix, query_start)
        if answer_start == -1:
            query = content[query_start:].strip()
        else:
            query = content[query_start:answer_start].strip()
        return query
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_query_from_content: %s", str(e))
        return ""

def update_input_span_events(args):
    if isinstance(args, tuple):
        return args[0].query_str if len(args) > 0 else ""


def update_output_span_events(results):
    if isinstance(results, list):
        output_arg_text = results[0].text
        if len(output_arg_text) > 100:
            output_arg_text = output_arg_text[:100] + "..."
        return output_arg_text


def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and hasattr(response, "raw"):
        if response.raw is not None:
            token_usage = response.raw.get("usage") if isinstance(response.raw, dict) else getattr(response.raw, "usage", None)
            if token_usage is not None:
                temperature = instance.__dict__.get("temperature", None)
                meta_dict.update({"temperature": temperature})
                if getattr(token_usage, "completion_tokens", None):
                    meta_dict.update({"completion_tokens": getattr(token_usage, "completion_tokens")})
                if getattr(token_usage, "prompt_tokens", None):
                    meta_dict.update({"prompt_tokens": getattr(token_usage, "prompt_tokens")})
                if getattr(token_usage, "total_tokens", None):
                    meta_dict.update({"total_tokens": getattr(token_usage, "total_tokens")})
    return meta_dict
